@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
 
-# Copyright (c) 2009-2014
+# Copyright (c) 2009-2017
 
 # Author(s):
 
@@ -23,17 +23,18 @@
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 """This module defines the image class. It overlaps largely the PIL library,
-but has the advandage of using masked arrays as pixel arrays, so that data
+but has the advantage of using masked arrays as pixel arrays, so that data
 arrays containing invalid values may be properly handled.
 """
 
 import logging
 import os
 import re
+from copy import deepcopy
 
-from PIL import Image as Pil
 import numpy as np
-
+import six
+from PIL import Image as Pil
 
 try:
     import numexpr as ne
@@ -42,6 +43,7 @@ except ImportError:
 
 logger = logging.getLogger(__name__)
 
+
 def ensure_dir(filename):
     """Checks if the dir of f exists, otherwise create it.
     """
@@ -51,9 +53,11 @@ def ensure_dir(filename):
 
 
 class UnknownImageFormat(Exception):
+
     """Exception to be raised when image format is unknown to pytroll-image"""
     pass
 
+
 def check_image_format(fformat):
     """Check that *fformat* is valid
     """
@@ -80,7 +84,9 @@ def check_image_format(fformat):
         raise UnknownImageFormat("Unknown image format '%s'." % fformat)
     return fformat
 
+
 class Image(object):
+
     """This class defines images. As such, it contains data of the different
     *channels* of the image (red, green, and blue for example). The *mode*
     tells if the channels define a black and white image ("L"), an rgb image
@@ -103,7 +109,7 @@ class Image(object):
     modes = ["L", "LA", "RGB", "RGBA", "YCbCr", "YCbCrA", "P", "PA"]
 
     def __init__(self, channels=None, mode="L", color_range=None,
-                 fill_value=None, palette=None):
+                 fill_value=None, palette=None, copy=True):
 
         self.channels = None
         self.mode = None
@@ -124,7 +130,13 @@ class Image(object):
 
         if(isinstance(channels, (tuple, list)) and
            len(channels) != len(re.findall("[A-Z]", mode))):
-            raise ValueError("Number of channels does not match mode.")
+            errmsg = ("Number of channels (" +
+                      "{n}) does not match mode {mode}.".format(
+                          n=len(channels), mode=mode))
+            raise ValueError(errmsg)
+
+        if copy and channels is not None:
+            channels = deepcopy(channels)
 
         if mode not in self.modes:
             raise ValueError("Unknown mode.")
@@ -137,13 +149,13 @@ class Image(object):
         if(color_range is not None and
            _is_list_of_pairs(color_range) and
            (channels is None or
-             len(color_range) != len(channels))):
+                len(color_range) != len(channels))):
             raise ValueError("Color_range length does not match number of "
                              "channels.")
 
         if(color_range is not None and
            (((mode == "L" or mode == "P") and not _is_pair(color_range)) and
-             (len(color_range) != len(re.findall("[A-Z]", mode))))):
+                (len(color_range) != len(re.findall("[A-Z]", mode))))):
             raise ValueError("Color_range does not match mode")
 
         self.mode = mode
@@ -165,10 +177,12 @@ class Image(object):
                     if color_range is not None:
                         color_min = color_range[i][0]
                         color_max = color_range[i][1]
+                        # Add data to image object as a channel
+                        #self._add_channel(chn, color_min, color_max)
                     else:
                         color_min = 0.0
                         color_max = 1.0
-
+                        # self.channels.append(np.ma.array(chn))
                     # Add data to image object as a channel
                     self._add_channel(chn, color_min, color_max)
 
@@ -177,7 +191,10 @@ class Image(object):
                         raise ValueError("Image channels must have the same"
                                          " shape.")
                 self.height = self.shape[0]
-                self.width = self.shape[1]
+                try:
+                    self.width = self.shape[1]
+                except IndexError:
+                    self.width = 0
             else:
                 raise ValueError("Image channels must all be arrays tuples.")
         elif channels is not None:
@@ -203,7 +220,6 @@ class Image(object):
     def _add_channel(self, chn, color_min, color_max):
         """Adds a channel to the image object
         """
-
         if isinstance(chn, np.ma.core.MaskedArray):
             chn_data = chn.data
             chn_mask = chn.mask
@@ -231,9 +247,11 @@ class Image(object):
             else:
                 final_data = chn.clip(0, 1) * np.iinfo(dtype).max
 
+            if np.issubdtype(dtype, np.integer):
+                final_data = np.round(final_data)
             channels.append(np.ma.array(final_data,
                                         dtype,
-                                        mask = np.ma.getmaskarray(chn)))
+                                        mask=np.ma.getmaskarray(chn)))
         if self.fill_value is not None:
             fill_value = [int(col * np.iinfo(dtype).max)
                           for col in self.fill_value]
@@ -249,13 +267,11 @@ class Image(object):
             raise RuntimeError("Channels-shape mismatch.")
         return self.channels == [] and self.shape == (0, 0)
 
-
     def show(self):
         """Display the image on screen.
         """
         self.pil_image().show()
 
-
     def pil_image(self):
         """Return a PIL image from the current image.
         """
@@ -335,18 +351,21 @@ class Image(object):
                 img = Pil.merge("RGBA", (pil_r, pil_g, pil_b, pil_a))
 
         else:
-            raise TypeError("Does not know how to use mode %s."%(self.mode))
+            raise TypeError("Does not know how to use mode %s." % (self.mode))
 
         return img
 
-    def save(self, filename, compression=6, fformat=None):
+    def save(self, filename, compression=6, fformat=None,
+             thumbnail_name=None, thumbnail_size=None):
         """Save the image to the given *filename*. For some formats like jpg
         and png, the work is delegated to :meth:`pil_save`, which doesn't
         support the *compression* option.
         """
-        self.pil_save(filename, compression, fformat)
+        self.pil_save(filename, compression, fformat,
+                      thumbnail_name, thumbnail_size)
 
-    def pil_save(self, filename, compression=6, fformat=None):
+    def pil_save(self, filename, compression=6, fformat=None,
+                 thumbnail_name=None, thumbnail_size=None):
         """Save the image to the given *filename* using PIL. For now, the
         compression level [0-9] is ignored, due to PIL's lack of support. See
         also :meth:`save`.
@@ -357,12 +376,54 @@ class Image(object):
         if self.is_empty():
             raise IOError("Cannot save an empty image")
 
-        ensure_dir(filename)
+        if isinstance(filename, (str, six.text_type)):
+            ensure_dir(filename)
 
         fformat = fformat or os.path.splitext(filename)[1][1:4]
         fformat = check_image_format(fformat)
 
-        self.pil_image().save(filename, fformat)
+        params = {}
+
+        if fformat == 'png':
+            # Take care of GeoImage.tags (if any).
+            params['pnginfo'] = self._pngmeta()
+
+        # JPEG images does not support transparency
+        if fformat == 'jpeg' and not self.fill_value:
+            self.fill_value = [0, 0, 0, 0]
+            logger.debug("No fill_value provided, setting it to 0.")
+
+        img = self.pil_image()
+        img.save(filename, fformat, **params)
+
+        if thumbnail_name is not None and thumbnail_size is not None:
+            img.thumbnail(thumbnail_size, Pil.ANTIALIAS)
+            img.save(thumbnail_name, fformat, **params)
+
+    def _pngmeta(self):
+        """It will return GeoImage.tags as a PNG metadata object.
+
+        Inspired by:
+        public domain, Nick Galbreath
+        http://blog.modp.com/2007/08/python-pil-and-png-metadata-take-2.html
+        """
+        reserved = ('interlace', 'gamma', 'dpi', 'transparency', 'aspect')
+
+        try:
+            tags = self.tags
+        except AttributeError:
+            tags = {}
+
+        # Undocumented class
+        from PIL import PngImagePlugin
+        meta = PngImagePlugin.PngInfo()
+
+        # Copy from tags to new dict
+        for k__, v__ in tags.items():
+            if k__ not in reserved:
+                meta.add_text(k__, v__, 0)
+
+        return meta
 
     def putalpha(self, alpha):
         """Adds an *alpha* channel to the current image, or replaces it with
@@ -375,7 +436,7 @@ class Image(object):
             raise ValueError("Alpha channel shape should match image shape")
 
         if not self.mode.endswith("A"):
-            self.convert(self.mode+"A")
+            self.convert(self.mode + "A")
         if not self.is_empty():
             self.channels[-1] = alpha
 
@@ -384,11 +445,10 @@ class Image(object):
 
         self._check_modes(("RGB", "RGBA"))
 
-
         (self.channels[0], self.channels[1], self.channels[2]) = \
-                          rgb2ycbcr(self.channels[0],
-                                    self.channels[1],
-                                    self.channels[2])
+            rgb2ycbcr(self.channels[0],
+                      self.channels[1],
+                      self.channels[2])
 
         if self.fill_value is not None:
             self.fill_value[0:3] = rgb2ycbcr(self.fill_value[0],
@@ -404,9 +464,9 @@ class Image(object):
         self._check_modes(("YCbCr", "YCbCrA"))
 
         (self.channels[0], self.channels[1], self.channels[2]) = \
-                          ycbcr2rgb(self.channels[0],
-                                    self.channels[1],
-                                    self.channels[2])
+            ycbcr2rgb(self.channels[0],
+                      self.channels[1],
+                      self.channels[2])
 
         if self.fill_value is not None:
             self.fill_value[0:3] = ycbcr2rgb(self.fill_value[0],
@@ -415,7 +475,6 @@ class Image(object):
 
         self.mode = mode
 
-
     def _to_p(self, mode):
         """Convert the image to P or PA mode.
         """
@@ -430,17 +489,19 @@ class Image(object):
             self._secondary_mode = self.mode
 
         palette = []
-        selfmask = reduce(np.ma.mask_or, [chn.mask for chn in chans])
-        new_chn = np.ma.zeros(self.shape, dtype = int)
+        selfmask = chans[0].mask
+        for chn in chans[1:]:
+            selfmask = np.ma.mask_or(selfmask, chn.mask)
+        new_chn = np.ma.zeros(self.shape, dtype=int)
         color_nb = 0
 
         for i in range(self.height):
             for j in range(self.width):
                 current_col = tuple([chn[i, j] for chn in chans])
                 try:
-                    (idx
-                     for idx in range(len(palette))
-                     if palette[idx] == current_col).next()
+                    next(idx
+                         for idx in range(len(palette))
+                         if palette[idx] == current_col)
                 except StopIteration:
                     idx = color_nb
                     palette.append(current_col)
@@ -456,9 +517,9 @@ class Image(object):
                 current_col = tuple(self.fill_value)
                 fill_alpha = []
             try:
-                (idx
-                 for idx in range(len(palette))
-                 if palette[idx] == current_col).next()
+                next(idx
+                     for idx in range(len(palette))
+                     if palette[idx] == current_col)
             except StopIteration:
                 idx = color_nb
                 palette.append(current_col)
@@ -497,7 +558,7 @@ class Image(object):
             new_chn = np.ma.array(np.interp(color_chan,
                                             np.arange(len(self.palette)),
                                             cdfs[i]),
-                                  mask = color_chan.mask)
+                                  mask=color_chan.mask)
             chans.append(new_chn)
 
         if self.fill_value is not None:
@@ -516,7 +577,6 @@ class Image(object):
 
         self.convert(mode)
 
-
     def _check_modes(self, modes):
         """Check that the image is in on of the given *modes*, raise an
         exception otherwise.
@@ -524,8 +584,7 @@ class Image(object):
         if not isinstance(modes, (tuple, list, set)):
             modes = [modes]
         if self.mode not in modes:
-            raise ValueError("Image not in suitable mode: %s"%modes)
-
+            raise ValueError("Image not in suitable mode: %s" % modes)
 
     def _l2rgb(self, mode):
         """Convert from L (black and white) to RGB.
@@ -537,7 +596,7 @@ class Image(object):
             self.fill_value = self.fill_value[:1] * 3 + self.fill_value[1:]
         if self.mode == "LA":
             self.channels[1], self.channels[3] = \
-            self.channels[3], self.channels[1]
+                self.channels[3], self.channels[1]
         self.mode = mode
 
     def _rgb2l(self, mode):
@@ -564,7 +623,6 @@ class Image(object):
 
         self.mode = mode
 
-
     def _ycbcr2l(self, mode):
         """Convert from YCbCr to L.
         """
@@ -589,11 +647,8 @@ class Image(object):
         if self.fill_value is not None:
             self.fill_value = [self.fill_value[0], 0, 0] + self.fill_value[1:]
 
-
         self.mode = mode
 
-
-
     def convert(self, mode):
         """Convert the current image to the given *mode*. See :class:`Image`
         for a list of available modes.
@@ -603,7 +658,7 @@ class Image(object):
 
         if mode not in ["L", "LA", "RGB", "RGBA",
                         "YCbCr", "YCbCrA", "P", "PA"]:
-            raise ValueError("Mode %s not recognized."%(mode))
+            raise ValueError("Mode %s not recognized." % (mode))
 
         if self.is_empty():
             self.mode = mode
@@ -653,19 +708,19 @@ class Image(object):
                       "YCbCr": self._from_p,
                       "L": self._from_p},
                 "PA": {"RGBA": self._from_p,
-                      "YCbCrA": self._from_p,
-                      "LA": self._from_p}}
+                       "YCbCrA": self._from_p,
+                       "LA": self._from_p}}
             try:
                 cases[self.mode][mode](mode)
             except KeyError:
                 raise ValueError("Conversion from %s to %s not implemented !"
-                                 %(self.mode,mode))
+                                 % (self.mode, mode))
 
-    def clip(self, channels = True):
+    def clip(self, channels=True):
         """Limit the values of the array to the default [0,1] range. *channels*
         says which channels should be clipped."""
         if not isinstance(channels, (tuple, list)):
-            channels = [channels]*len(self.channels)
+            channels = [channels] * len(self.channels)
 
         for i in range(len(self.channels)):
             if channels[i]:
@@ -704,19 +759,19 @@ class Image(object):
         i = 0
         for chn in self.channels:
             if zoom[0]:
-                chn = chn.repeat([factor[0]] * chn.shape[0], axis = 0)
+                chn = chn.repeat([factor[0]] * chn.shape[0], axis=0)
             else:
                 chn = chn[[idx * factor[0]
-                           for idx in range(self.height / factor[0])],
+                           for idx in range(int(self.height / factor[0]))],
                           :]
             if zoom[1]:
                 self.channels[i] = chn.repeat([factor[1]] * chn.shape[1],
-                                              axis = 1)
+                                              axis=1)
             else:
                 self.channels[i] = chn[:,
                                        [idx * factor[1]
-                                        for idx in range(self.width /
-                                                         factor[1])]]
+                                        for idx in range(int(self.width /
+                                                             factor[1]))]]
 
             i = i + 1
 
@@ -734,7 +789,7 @@ class Image(object):
 
         if luminance.shape != self.channels[0].shape:
             if ((luminance.shape[0] * 1.0 / luminance.shape[1]) ==
-                 (self.channels[0].shape[0] * 1.0 / self.channels[0].shape[1])):
+                    (self.channels[0].shape[0] * 1.0 / self.channels[0].shape[1])):
                 if luminance.shape[0] > self.channels[0].shape[0]:
                     self.resize(luminance.shape)
                 else:
@@ -752,15 +807,20 @@ class Image(object):
             self.channels[0] = luminance
             self.convert(mode)
 
-    def enhance(self, inverse=False, gamma=1.0, stretch="no"):
+    def enhance(self, inverse=False, gamma=1.0, stretch="no",
+                stretch_parameters=None, **kwargs):
         """Image enhancement function. It applies **in this order** inversion,
         gamma correction, and stretching to the current image, with parameters
         *inverse* (see :meth:`Image.invert`), *gamma* (see
         :meth:`Image.gamma`), and *stretch* (see :meth:`Image.stretch`).
         """
         self.invert(inverse)
+        if stretch_parameters is None:
+            stretch_parameters = {}
+
+        stretch_parameters.update(kwargs)
+        self.stretch(stretch, **stretch_parameters)
         self.gamma(gamma)
-        self.stretch(stretch)
 
     def gamma(self, gamma=1.0):
         """Apply gamma correction to the channels of the image. If *gamma* is a
@@ -770,46 +830,42 @@ class Image(object):
         are several channels in the image. The behaviour of :func:`gamma` is
         undefined outside the normal [0,1] range of the channels.
         """
-        if not isinstance(gamma, (int, long, float)):
-            if(not isinstance(gamma, (tuple, list, set)) or
-               not _areinstances(gamma, (int, long, float))):
-                raise TypeError("Gamma should be a real number, or an iterable "
-                                "of real numbers.")
 
         if(isinstance(gamma, (list, tuple, set)) and
            len(gamma) != len(self.channels)):
             raise ValueError("Number of channels and gamma components differ.")
-
-        if gamma < 0:
-            raise ValueError("Gamma correction must be a positive number.")
-
-        if gamma == 1.0:
-            return
         if isinstance(gamma, (tuple, list)):
             gamma_list = list(gamma)
         else:
             gamma_list = [gamma] * len(self.channels)
         for i in range(len(self.channels)):
+            gamma = float(gamma_list[i])
+            if gamma < 0:
+                raise ValueError("Gamma correction must be a positive number.")
+            logger.debug("Applying gamma %f", gamma)
+            if gamma == 1.0:
+                continue
+
             if isinstance(self.channels[i], np.ma.core.MaskedArray):
                 if ne:
                     self.channels[i] = np.ma.array(
                         ne.evaluate("data ** (1.0 / gamma)",
                                     local_dict={"data": self.channels[i].data,
-                                                'gamma': gamma_list[i]}),
+                                                'gamma': gamma}),
                         mask=self.channels[i].mask,
                         copy=False)
                 else:
                     self.channels[i] = np.ma.array(self.channels[i].data **
-                                                   (1.0 / gamma_list[i]),
+                                                   (1.0 / gamma),
                                                    mask=self.channels[i].mask,
                                                    copy=False)
             else:
                 self.channels[i] = np.where(self.channels[i] >= 0,
                                             self.channels[i] **
-                                            (1.0 / gamma_list[i]),
+                                            (1.0 / gamma),
                                             self.channels[i])
 
-    def stretch(self, stretch="crude", **kwarg):
+    def stretch(self, stretch="crude", **kwargs):
         """Apply stretching to the current image. The value of *stretch* sets
         the type of stretching applied. The values "histogram", "linear",
         "crude" (or "crude-stretch") perform respectively histogram
@@ -821,58 +877,59 @@ class Image(object):
         range [0.0,1.0].
         """
 
+        logger.debug("Applying stretch %s with parameters %s",
+                     stretch, str(kwargs))
+
         ch_len = len(self.channels)
         if self.mode.endswith("A"):
             ch_len -= 1
 
-
         if((isinstance(stretch, tuple) or
-             isinstance(stretch, list))):
+                isinstance(stretch, list))):
             if len(stretch) == 2:
                 for i in range(ch_len):
-                    self.stretch_linear(i, cutoffs=stretch, **kwarg)
+                    self.stretch_linear(i, cutoffs=stretch, **kwargs)
             else:
-                raise ValueError("Stretch tuple must have exactly two elements")
+                raise ValueError(
+                    "Stretch tuple must have exactly two elements")
         elif stretch == "linear":
             for i in range(ch_len):
-                self.stretch_linear(i, **kwarg)
+                self.stretch_linear(i, **kwargs)
         elif stretch == "histogram":
             for i in range(ch_len):
-                self.stretch_hist_equalize(i, **kwarg)
+                self.stretch_hist_equalize(i, **kwargs)
         elif stretch in ["crude", "crude-stretch"]:
             for i in range(ch_len):
-                self.crude_stretch(i, **kwarg)
+                self.crude_stretch(i, **kwargs)
         elif stretch in ["log", "logarithmic"]:
             for i in range(ch_len):
-                self.stretch_logarithmic(i, **kwarg)
+                self.stretch_logarithmic(i, **kwargs)
         elif stretch == "no":
             return
         elif isinstance(stretch, str):
-            raise ValueError("Stretching method %s not recognized."%stretch)
+            raise ValueError("Stretching method %s not recognized." % stretch)
         else:
             raise TypeError("Stretch parameter must be a string or a tuple.")
 
-
     def invert(self, invert=True):
-        """Inverts all the channels of a image according to *invert*. If invert
-        is a tuple or a list, elementwise invertion is performed, otherwise all
-        channels are inverted if *invert* is true (default).
-        """
-        if(isinstance(invert, (tuple, list, set)) and
-           len(self.channels) != len(invert)):
-            raise ValueError("Number of channels and invert components differ.")
+        """Inverts all the channels of a image according to *invert*. If invert is a tuple or a list, elementwise
+        invertion is performed, otherwise all channels are inverted if *invert* is true (default).
 
-        if isinstance(invert, (tuple, list, set)):
-            i = 0
-            for chn in self.channels:
+        Note: 'Inverting' means that black becomes white, and vice-versa, not that the values are negated !
+        """
+        if(isinstance(invert, (tuple, list)) and
+           len(self.channels) != len(invert)):
+            raise ValueError(
+                "Number of channels and invert components differ.")
+
+        logger.debug("Applying invert with parameters %s", str(invert))
+        if isinstance(invert, (tuple, list)):
+            for i, chn in enumerate(self.channels):
                 if invert[i]:
-                    self.channels[i] = 1.0 - chn
-                i = i + 1
+                    self.channels[i] = 1 - chn
         elif invert:
-            i = 0
-            for chn in self.channels:
-                self.channels[i] = 1.0 - chn
-                i = i + 1
+            for i, chn in enumerate(self.channels):
+                self.channels[i] = 1 - chn
 
     def stretch_hist_equalize(self, ch_nb):
         """Stretch the current image's colors by performing histogram
@@ -891,9 +948,9 @@ class Image(object):
 
         carr = arr.compressed()
 
-        cdf = np.arange(0.0, 1.0, 1/nwidth)
+        cdf = np.arange(0.0, 1.0, 1 / nwidth)
         logger.debug("Make histogram bins having equal amount of data, " +
-                  "using numpy percentile function:")
+                     "using numpy percentile function:")
         bins = np.percentile(carr, list(cdf * 100))
 
         res = np.ma.empty_like(arr)
@@ -902,7 +959,6 @@ class Image(object):
 
         self.channels[ch_nb] = res
 
-
     def stretch_logarithmic(self, ch_nb, factor=100.):
         """Move data into range [1:factor] and do a normalized logarithmic
         enhancement.
@@ -910,7 +966,7 @@ class Image(object):
         logger.debug("Perform a logarithmic contrast stretch.")
         if ((self.channels[ch_nb].size ==
              np.ma.count_masked(self.channels[ch_nb])) or
-              (self.channels[ch_nb].min() == self.channels[ch_nb].max())):
+                (self.channels[ch_nb].min() == self.channels[ch_nb].max())):
             logger.warning("Nothing to stretch !")
             return
 
@@ -919,9 +975,9 @@ class Image(object):
         arr = self.channels[ch_nb]
         b__ = float(crange[1] - crange[0]) / np.log(factor)
         c__ = float(crange[0])
-        slope = (factor-1.)/float(arr.max() - arr.min())
-        arr = 1. + (arr - arr.min())*slope
-        arr = c__ + b__*np.log(arr)
+        slope = (factor - 1.) / float(arr.max() - arr.min())
+        arr = 1. + (arr - arr.min()) * slope
+        arr = c__ + b__ * np.log(arr)
         self.channels[ch_nb] = arr
 
     def stretch_linear(self, ch_nb, cutoffs=(0.005, 0.005)):
@@ -941,23 +997,21 @@ class Image(object):
 
         logger.debug("Calculate the histogram percentiles: ")
         logger.debug("Left and right percentiles: " +
-                  str(cutoffs[0]*100) + " " + str(cutoffs[1]*100))
+                     str(cutoffs[0] * 100) + " " + str(cutoffs[1] * 100))
 
-        left, right = np.percentile(carr, [cutoffs[0]*100,
-                                           100. - cutoffs[1]*100])
+        left, right = np.percentile(
+            carr, [cutoffs[0] * 100, 100. - cutoffs[1] * 100])
 
         delta_x = (right - left)
         logger.debug("Interval: left=%f, right=%f width=%f",
-                   left, right, delta_x)
+                     left, right, delta_x)
 
         if delta_x > 0.0:
             self.channels[ch_nb] = np.ma.array((arr - left) / delta_x,
-                                               mask = arr.mask)
+                                               mask=arr.mask)
         else:
-            self.channels[ch_nb] = np.ma.zeros(arr.shape)
             logger.warning("Unable to make a contrast stretch!")
 
-
     def crude_stretch(self, ch_nb, min_stretch=None, max_stretch=None):
         """Perform simple linear stretching (without any cutoff) on the channel
         *ch_nb* of the current image and normalize to the [0,1] range."""
@@ -967,8 +1021,13 @@ class Image(object):
         if max_stretch is None:
             max_stretch = self.channels[ch_nb].max()
 
+        if isinstance(min_stretch, (list, tuple)):
+            min_stretch = min_stretch[ch_nb]
+        if isinstance(max_stretch, (list, tuple)):
+            max_stretch = max_stretch[ch_nb]
+
         if((not self.channels[ch_nb].mask.all()) and
-            max_stretch - min_stretch > 0):
+                abs(max_stretch - min_stretch) > 0):
             stretched = self.channels[ch_nb].data.astype(np.float)
             stretched -= min_stretch
             stretched /= max_stretch - min_stretch
@@ -988,7 +1047,9 @@ class Image(object):
         if self.mode != img.mode:
             raise ValueError("Cannot merge image of different modes.")
 
-        selfmask = reduce(np.ma.mask_or, [chn.mask for chn in self.channels])
+        selfmask = self.channels[0].mask
+        for chn in self.channels[1:]:
+            selfmask = np.ma.mask_or(selfmask, chn.mask)
 
         for i in range(len(self.channels)):
             self.channels[i] = np.ma.where(selfmask,
@@ -1043,11 +1104,19 @@ class Image(object):
             dst.channels[i][outa == 0] = 0
         dst.channels[3] = outa
 
+    def _repr_png_(self):
+        import io
+        b = io.BytesIO()
+        self.save(b, fformat="png")
+        return b.getvalue()
+
+
 def _areinstances(the_list, types):
     """Check if all the elements of the list are of given type.
     """
     return all([isinstance(item, types) for item in the_list])
 
+
 def _is_pair(item):
     """Check if an item is a pair (tuple of size 2).
     """
@@ -1056,6 +1125,7 @@ def _is_pair(item):
             not isinstance(item[0], (list, tuple, set)) and
             not isinstance(item[1], (list, tuple, set)))
 
+
 def _is_list_of_pairs(the_list):
     """Check if a list contains only pairs.
     """
@@ -1075,6 +1145,7 @@ def ycbcr2rgb(y__, cb_, cr_):
 
     return r__, g__, b__
 
+
 def rgb2ycbcr(r__, g__, b__):
     """Convert the three RGB channels to YCbCr."""
 
@@ -1086,4 +1157,3 @@ def rgb2ycbcr(r__, g__, b__):
     cr_ = 1. / (2 * (1 - kr_)) * (r__ - y__)
 
     return y__, cb_, cr_
-
